#include "main.h"
#include <commctrl.h>
#include "display.h"
#include "resource.h"


#define	GENERAL_REGISTERS_NUMBER		8
#define	FPU_REGISTERS_NUMBER			8
#define	MMX_REGISTERS_NUMBER			8
#define SEG_CTRL_REGISTERS_NUMBER		8

#ifdef X86_64
#define CR_TR_REGISTERS_NUMBER			6
#else
#define CR_TR_REGISTERS_NUMBER			5
#endif

#define DEBUG_REGISTERS_NUMBER			6
#define	FLAGS_NUMBER					9


#define SEG_REGISTERS_NUMBER			6

#define	MAIN_VIEWS_NUMBER				6


extern	x86_emul_context	context;
extern sDEFAULT *general_register_values[];
extern long double *fpu_register_values[];
extern sQWORD *mmx_register_values[];
extern sDEFAULT *control_register_values[];
extern sDEFAULT *debug_register_values[];
extern sADDR stack_segment_base;

extern	int emulator_enable;


unsigned __int32 *seg_ctrl_register_values[] = {
(unsigned __int32*)&context.R_es, (unsigned __int32*)&context.R_cs,
(unsigned __int32*)&context.R_ss, (unsigned __int32*)&context.R_ds,
(unsigned __int32*)&context.R_fs, (unsigned __int32*)&context.R_gs, 
&context.R_ip, &context.R_flags};

HINSTANCE	plugin_instance;
HACCEL		accelerator;

	//Register names and flags mask

char	*reg_general[] = {"EAX", "ECX", "EDX", "EBX", "ESP", "EBP", "ESI", "EDI"};
char	*reg_fpu[] = {"ST0", "ST1", "ST2", "ST3", "ST4", "ST5", "ST6", "ST7"};
char	*reg_mmx[] = {"MM0", "MM1", "MM2", "MM3", "MM4", "MM5", "MM6", "MM7"};

#ifdef X86_64
char	*reg_ctrl[] = {"CR0", "CR2", "CR3", "CR4", "CR8", "TR"};
#else
char	*reg_ctrl[] = {"CR0", "CR2", "CR3", "CR4", "TR"};
#endif

char	*reg_dbg[] = {"DR0", "DR1", "DR2", "DR3", "DR6", "DR7"};
char	*reg_segment_ctrl[] = {"ES", "CS", "SS", "DS", "FS", "GS", "EIP", "EFLAGS"};
char	*flags[] = {"C", "P", "A", "Z", "S", "T", "I", "D", "O"};

unsigned __int32 flags_mask[] = {FLAG_C, FLAG_P, FLAG_A, FLAG_Z, FLAG_S,
								 FLAG_T, FLAG_I, FLAG_D, FLAG_O};



	//Additional information types

char	*main_views[] = {"General purpose registers",
	                     "FPU registers",
	                     "MMX registers",
	                     "Task and control registers",
	                     "Debug registers",
	                     "breakpoints"};

	//Emulator window data
char *wndclass = "IDAx86_wndclass";

WNDCLASSEX	wc;
MSG			msg;


HFONT					dialog_font;
WNDPROC					old_stack_wndproc;
sADDR					stack_view_addr;


int QXhextostr0A_16(short num, char *buff, unsigned long sign);
int QXhextostr0A_32(long num, char *buff, unsigned long sign);
int QXhextostr0A_64(__int64 num, char *buff, unsigned long sign);

bool QXstrtohex_16(char *hex, short *out);
bool QXstrtohex_32(char *hex, long *out);
bool QXstrtohex_64(char *hex, __int64 *out);

char* QXstrcpyt(char *buff, char *str, char delimiter);
char* QXstrcpytn(char *buff, char *str, char delimiter, int len);

void InitComboBoxes(HWND hWnd);
void InitListViews(HWND hWnd);
void InitStackView(HWND hWnd);

LRESULT emul_window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT stack_window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

/* Functions for refresh individual types
   of information in emulator window */

void refresh_listview(HWND hWnd, listview_content *LV);
void refresh_stack(HWND hWnd);
void refresh_status(HWND hWnd);

bool inputdlg_window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);


bool OpenDialogExecute(HINSTANCE hInstance, HWND owner, char *filter, char *filename);
bool SaveDialogExecute(HINSTANCE hInstance, HWND owner, char *filter, char *filename);

#define		OPERATION_READ		0x01
#define		OPERATION_WRITE		0x02

LRESULT dumpdialog_window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);


bool f_in_16bit(void *out, sDEFAULT number, char *data);
bool f_in_32bit(void *out, sDEFAULT number, char *data);
bool f_in_64bit(void *out, sDEFAULT number, char *data);
bool f_in_fpu(void *out, sDEFAULT number, char *data);
bool f_in_bp(void *out, sDEFAULT number, char *data);

void f_out_16bit(void *data, sDEFAULT number, char *out);
void f_out_32bit(void *data, sDEFAULT number, char *out);
void f_out_64bit(void *data, sDEFAULT number, char *out);
void f_out_fpu(void *data, sDEFAULT number, char *out);
void f_out_flags(void *data, sDEFAULT number, char *out);
void f_out_bp(void *data, sDEFAULT number, char *out);

void f_GP_descr(void *data, sDEFAULT number, char *out);
void f_FPU_descr(void *data, sDEFAULT number, char *out);
void f_MMX_descr(void *data, sDEFAULT number, char *out);
void f_CR_TR_descr(void *data, sDEFAULT number, char *out);
void f_DR_descr(void *data, sDEFAULT number, char *out);
void f_SEG_descr(void *data, sDEFAULT number, char *out);
void f_FLAGS_descr(void *data, sDEFAULT number, char *out);
void f_bp_descr(void *data, sDEFAULT number, char *out);

void f_dblclick_regview(HWND hWnd, item_properties *item, int item_number);
void f_dblclick_flags(HWND hWnd, item_properties *item, int item_number);
void f_dblclick_bp(HWND hWnd, item_properties *item, int item_number);

item_properties GP_regs[] = {
{&context.R_ax, &f_GP_descr, &f_in_32bit, &f_out_32bit},
{&context.R_cx, &f_GP_descr, &f_in_32bit, &f_out_32bit},
{&context.R_dx, &f_GP_descr, &f_in_32bit, &f_out_32bit},
{&context.R_bx, &f_GP_descr, &f_in_32bit, &f_out_32bit},
{&context.R_sp, &f_GP_descr, &f_in_32bit, &f_out_32bit},
{&context.R_bp, &f_GP_descr, &f_in_32bit, &f_out_32bit},
{&context.R_si, &f_GP_descr, &f_in_32bit, &f_out_32bit},
{&context.R_di, &f_GP_descr, &f_in_32bit, &f_out_32bit}};

item_properties FPU_regs[] = {
{&context.R_st0, &f_FPU_descr, &f_in_fpu, &f_out_fpu},
{&context.R_st1, &f_FPU_descr, &f_in_fpu, &f_out_fpu},
{&context.R_st2, &f_FPU_descr, &f_in_fpu, &f_out_fpu},
{&context.R_st3, &f_FPU_descr, &f_in_fpu, &f_out_fpu},
{&context.R_st4, &f_FPU_descr, &f_in_fpu, &f_out_fpu},
{&context.R_st5, &f_FPU_descr, &f_in_fpu, &f_out_fpu},
{&context.R_st6, &f_FPU_descr, &f_in_fpu, &f_out_fpu},
{&context.R_st7, &f_FPU_descr, &f_in_fpu, &f_out_fpu}};

item_properties MMX_regs[] = {
{&context.R_MM0, &f_MMX_descr, &f_in_64bit, &f_out_64bit},
{&context.R_MM1, &f_MMX_descr, &f_in_64bit, &f_out_64bit},
{&context.R_MM2, &f_MMX_descr, &f_in_64bit, &f_out_64bit},
{&context.R_MM3, &f_MMX_descr, &f_in_64bit, &f_out_64bit},
{&context.R_MM4, &f_MMX_descr, &f_in_64bit, &f_out_64bit},
{&context.R_MM5, &f_MMX_descr, &f_in_64bit, &f_out_64bit},
{&context.R_MM6, &f_MMX_descr, &f_in_64bit, &f_out_64bit},
{&context.R_MM7, &f_MMX_descr, &f_in_64bit, &f_out_64bit}};

#ifdef X86_64
item_properties CR_TR_regs[] = {
{&context.R_cr0, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_cr2, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_cr3, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_cr4, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_cr8, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_tr, &f_CR_TR_descr, &f_in_16bit, &f_out_16bit}};
#else
item_properties CR_TR_regs[] = {
{&context.R_cr0, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_cr2, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_cr3, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_cr4, &f_CR_TR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_tr, &f_CR_TR_descr, &f_in_16bit, &f_out_16bit}};
#endif

item_properties DR_regs[] = {
{&context.R_dr0, &f_DR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_dr1, &f_DR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_dr2, &f_DR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_dr3, &f_DR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_dr6, &f_DR_descr, &f_in_32bit, &f_out_32bit},
{&context.R_dr7, &f_DR_descr, &f_in_32bit, &f_out_32bit}};

item_properties SEG_regs[] = {
{&context.R_es, &f_SEG_descr, &f_in_16bit, &f_out_16bit},
{&context.R_cs, &f_SEG_descr, &f_in_16bit, &f_out_16bit},
{&context.R_ss, &f_SEG_descr, &f_in_16bit, &f_out_16bit},
{&context.R_ds, &f_SEG_descr, &f_in_16bit, &f_out_16bit},
{&context.R_fs, &f_SEG_descr, &f_in_16bit, &f_out_16bit},
{&context.R_gs, &f_SEG_descr, &f_in_16bit, &f_out_16bit},
{&context.R_ip, &f_SEG_descr, &f_in_32bit, &f_out_32bit},
{&context.R_flags, &f_SEG_descr, &f_in_32bit, &f_out_32bit}};


item_properties FLAGS[] = {
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags},
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags},
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags},
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags},
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags},
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags},
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags},
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags},
{&context.R_flags, &f_FLAGS_descr, NULL, &f_out_flags}};

item_properties BREAKPOINTS[] = {
{&context.breakpoints[0], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[1], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[2], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[3], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[4], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[5], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[6], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[7], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[8], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[9], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[10], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[11], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[12], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[13], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[14], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[15], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[16], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[17], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[18], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[19], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[20], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[21], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[22], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[23], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[24], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[25], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[26], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[27], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[28], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[29], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[30], &f_bp_descr, &f_in_bp, &f_out_bp},
{&context.breakpoints[31], &f_bp_descr, &f_in_bp, &f_out_bp}
};


listview_properties GP_listview = {
GENERAL_REGISTERS_NUMBER, GP_regs, "Name", "Value", 25, 75, &f_dblclick_regview
};

listview_properties FPU_listview = {
FPU_REGISTERS_NUMBER, FPU_regs, "Name", "Value", 25, 75, &f_dblclick_regview
};

listview_properties MMX_listview = {
MMX_REGISTERS_NUMBER, MMX_regs, "Name", "Value", 25, 75, &f_dblclick_regview
};

listview_properties CR_TR_listview = {
CR_TR_REGISTERS_NUMBER, CR_TR_regs, "Name", "Value", 25, 75, &f_dblclick_regview
};

listview_properties DR_listview = {
DEBUG_REGISTERS_NUMBER, DR_regs, "Name", "Value", 25, 75, &f_dblclick_regview
};

listview_properties BP_listview = {
0, BREAKPOINTS, "Number", "Type", 25, 75, &f_dblclick_bp
};

listview_properties SEG_listview = {
SEG_CTRL_REGISTERS_NUMBER, SEG_regs, "Name", "Value", 33, 66, &f_dblclick_regview
};

listview_properties FLAGS_listview = {
FLAGS_NUMBER, FLAGS, "Flag", "Value", 40, 60, &f_dblclick_flags
};


listview_content lv1 = {
ListView1, &GP_listview, 0
};

listview_content lv2 = {
ListView2, &FPU_listview, 0
};

listview_content seglv = {
SGListView, &SEG_listview, 0
};

listview_content flagslv = {
FlagsListView, &FLAGS_listview, 0
};

listview_properties *prop_table[] = {
&GP_listview, &FPU_listview, &MMX_listview, &CR_TR_listview, &DR_listview, &BP_listview
};

listview_content *LVs[] = {&lv1, &lv2, &seglv, &flagslv};

HWND create_emulator_window()
{HWND hDlg;
 HICON icon;
 int i;

 plugin_instance = GetModuleHandle("IDAx86.plw");
 icon = LoadIcon(NULL, IDI_APPLICATION);
 accelerator = LoadAccelerators(plugin_instance, MAKEINTRESOURCE(IDR_ACCELERATOR1));
 dialog_font = (HFONT)GetStockObject(SYSTEM_FIXED_FONT);

 wc.cbSize = sizeof(WNDCLASSEX);
 wc.style = CS_HREDRAW + CS_VREDRAW;
 wc.lpfnWndProc = &emul_window_proc;
 wc.cbClsExtra = NULL;
 wc.cbWndExtra = DLGWINDOWEXTRA;
 wc.hInstance = plugin_instance;
 wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
 wc.lpszMenuName = "MENU_0";
 wc.lpszClassName = wndclass;
 wc.hIcon = icon;
 wc.hIconSm = icon;
 wc.hCursor = LoadCursor(NULL, IDC_ARROW);
 RegisterClassEx(&wc);

 hDlg = CreateDialogParam(plugin_instance, MAKEINTRESOURCE(EmulDialog), NULL, NULL, NULL);
 if (hDlg != 0)
   {InitComboBoxes(hDlg);
    InitListViews(hDlg);
	InitStackView(hDlg);
	return hDlg;}
	 
  else {MessageBox(NULL, "Cannot create emulator window", "ERROR!!!", MB_ICONERROR);
        return 0;}}

void show_emulator_window(HWND hDlg)
{ShowWindow(hDlg, SW_SHOWNORMAL);
 UpdateWindow(hDlg);
 refresh_all(hDlg);
 
 while (GetMessage(&msg, NULL, 0, 0))
   {if (!TranslateAccelerator(hDlg, accelerator, &msg))
      TranslateMessage(&msg);
    if (emulator_enable)
	  {if (msg.hwnd == hDlg || IsChild(hDlg, msg.hwnd))
         DispatchMessage(&msg);}
	else DispatchMessage(&msg);}}




void destroy_emulator_window(HWND hDlg)
{DestroyWindow(hDlg);}


LRESULT emul_window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{dumpdialog_params dmp;
 bool result;
 sDEFAULT old_esp;
 DWORD thread_id;
 char buffer[128];
 int i, index;
 listview_properties *lv_prop;

 HANDLE hFile;
 DWORD hFileSize, f_operation;
	
 switch (uMsg)
  {case WM_COMMAND :
     if (lParam)
       {switch (wParam&0xFFFF)
          {case ViewCombo1 :
	         if (wParam>>16 == CBN_SELCHANGE)
			   {lv1.prop = prop_table[SendDlgItemMessage(hWnd, ViewCombo1, CB_GETCURSEL, NULL, NULL)];
				refresh_listview(hWnd, &lv1);}
			   
             break;

           case ViewCombo2 :
             if (wParam>>16 == CBN_SELCHANGE)
			   {lv2.prop = prop_table[SendDlgItemMessage(hWnd, ViewCombo2, CB_GETCURSEL, NULL, NULL)];
				refresh_listview(hWnd, &lv2);}
             break;

		   case StatusButton :
			 if (wParam>>16 == BN_CLICKED)
			   {emulator_enable = 0;
		        refresh_status(hWnd);}
		   break;
				

           }
     
	    }
	 else
	   {switch (wParam&0xFFFF)
	      {case ID_FILE_HIDE :
             ShowWindow(hWnd, SW_HIDE);
             PostQuitMessage(NULL);
			 break;
	   
           case ID_DEBUG_ADJUSTIPTOCURSOR :
			 change_eip();
		     refresh_listview(hWnd, &seglv);
		     break;

		   case ID_DEBUG_ADJUSTCURSORTOIP :
		     change_cursor_pos();
		     break;

	       case ID_DEBUG_TRACEINTO :
			 result = trace_into(context.R_ip);
		     if (result) refresh_all(hWnd);
			 break;

		   case ID_DEBUG_STEPOVER :
		     if (!emulator_enable)
		       {emulator_enable = 1;
                refresh_status(hWnd);

                CreateThread(NULL, 0x10000, (LPTHREAD_START_ROUTINE)&step_over,
				             (LPVOID)get_cursor_pos(), THREAD_PRIORITY_HIGHEST, &thread_id);}
			 break;


		   case ID_DEBUG_RUNTOCURSOR :
             
		     if (!emulator_enable)
		       {emulator_enable = 1;
                refresh_status(hWnd);

                CreateThread(NULL, 0x10000, (LPTHREAD_START_ROUTINE)&run_to_cursor,
				             (LPVOID)get_cursor_pos(), THREAD_PRIORITY_HIGHEST, &thread_id);}
		     break;

	   case ID_DEBUG_IGNOREITEM :
	    context.R_ip += item_size(context.R_ip);
            refresh_listview(hWnd, &seglv);
	    change_cursor_pos();
	    break;

           case ID_DEBUG_RUN :
		     if (!emulator_enable)
		       {emulator_enable = 1;
                refresh_status(hWnd);

                CreateThread(NULL, 0x10000, (LPTHREAD_START_ROUTINE)&e_run,
				             (LPVOID)get_cursor_pos(), THREAD_PRIORITY_HIGHEST, &thread_id);}
		     break;


           case ID_DEBUG_EXECUTETILLRETURN :
		     if (!emulator_enable)
		       {emulator_enable = 1;
                refresh_status(hWnd);

                CreateThread(NULL, 0x10000, (LPTHREAD_START_ROUTINE)&execute_till_return,
				             (LPVOID)get_cursor_pos(), THREAD_PRIORITY_HIGHEST, &thread_id);}
		     break;


	   case ID_DEBUG_RESETREGISTERS :
            for (i = 0; i < GENERAL_REGISTERS_NUMBER; i++)
	     *general_register_values[i] = 0;
            for (i = 0; i < FPU_REGISTERS_NUMBER; i++)
	     *fpu_register_values[i] = 0;
            for (i = 0; i < CONTROL_REGISTERS_NUMBER; i++)
	     *control_register_values[i] = 0;
            for (i = 0; i < DEBUG_REGISTERS_NUMBER; i++)
	     *debug_register_values[i] = 0;
	    
	    context.R_flags = FLAG_I;
	    context.R_cs = 0x1B;
	    context.R_ds = context.R_ss = context.R_es = 0x23;
	    context.R_fs = 0x38;
	    context.R_sp = stack_segment_base + 0xFFC4;
	    
	    refresh_all(hWnd);
	    break;

	   case ID_DEBUG_DELETEALLBREAKPOINTS :
	    memset(&context.breakpoints, 0, sizeof(context.breakpoints));
            context.breakpoints_number = 0;
	    context.breakpoints[0].type = ~0;
	    
            refresh_all(hWnd);
	    break;

	   case ID_DEBUG_FULLRESETEMULATOR :
	    delete_stack_segment();
	    delete_peb_segment();
	    
	    reset_configuration();
	    
	    refresh_all(hWnd);
	    break;

	   case ID_SEGMENTS_CREATESTACKSEGMENT :
	    create_stack_segment();
	    refresh_stack(hWnd);
	    break;

           case ID_SEGMENTS_CREATEPEBSEGMENT :
		     create_peb_segment();
		     break;

           case ID_SEGMENTS_DELETESTACKSEGMENT :
		     delete_stack_segment();
		     refresh_stack(hWnd);
		     break;
			
           case ID_SEGMENTS_DELETEPEBSEGMENT :
		     delete_peb_segment();
		     break;

           case ID_FILE_DUMPIDABASE :
			 dmp.title = "Save dump from database";
			 dmp.operation = OPERATION_WRITE;
			 DialogBoxParam(plugin_instance, MAKEINTRESOURCE(DumpDialog), hWnd, (DLGPROC)&dumpdialog_window_proc, (LPARAM)&dmp);
		     break;

		   case ID_FILE_LOADDUMP :
		     dmp.title = "Load dump to database";
			 dmp.operation = OPERATION_READ;
			 DialogBoxParam(plugin_instance, MAKEINTRESOURCE(DumpDialog), hWnd, (DLGPROC)&dumpdialog_window_proc, (LPARAM)&dmp);
		     break;
 
		   case ID_FILE_SAVECONTEXTTOFILE :
		     f_operation = 0;
		     SaveDialogExecute(plugin_instance, hWnd,
				               "Emulator context files(*.x86c)\0*.x86c\0All files(*.*)\0*.*\0\0",
							   buffer);
			 hFile = CreateFile(buffer, GENERIC_WRITE, NULL, NULL,
				                OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
			 if (hFile != INVALID_HANDLE_VALUE)
			   {WriteFile(hFile, &context, sizeof(context), &hFileSize, NULL);
			    SetEndOfFile(hFile);
				CloseHandle(hFile);}

			 if (hFileSize != sizeof(context))
			   MessageBox(hWnd, "Cannot save context to this file", buffer, MB_ICONERROR);
			 
			 break;

		   case ID_FILE_LOADCONTEXTFROMFILE :
		     OpenDialogExecute(plugin_instance, hWnd,
				               "Emulator context files(*.x86c)\0*.x86c\0All files(*.*)\0*.*\0\0",
							   buffer);
			 hFile = CreateFile(buffer, GENERIC_READ, FILE_SHARE_READ, NULL,
				                OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
			 if (hFile != INVALID_HANDLE_VALUE)
			   {ReadFile(hFile, &context, sizeof(context), &hFileSize, NULL);
				CloseHandle(hFile);}

			 if (hFileSize != sizeof(context))
			   MessageBox(hWnd, "Cannot load context from this file", buffer, MB_ICONERROR);
			 else
			   refresh_all(hWnd);
			 break;

	       }}
	 break;

	 case WM_NOTIFY :
	   if ((((NMHDR*)lParam)->idFrom == ListView1 ||
		   ((NMHDR*)lParam)->idFrom == ListView2 ||
		   ((NMHDR*)lParam)->idFrom == SGListView ||
		   ((NMHDR*)lParam)->idFrom == FlagsListView) &&
		   (((NMHDR*)lParam)->code == NM_DBLCLK || ((NMHDR*)lParam)->code == NM_RETURN))
	     {for (i = 0; i < sizeof(LVs)/sizeof(listview_content*); i++)
		    if (LVs[i]->ID == ((NMHDR*)lParam)->idFrom)
			  break;
	      lv_prop = LVs[i]->prop;
		  index = SendDlgItemMessage(hWnd, LVs[i]->ID, LVM_GETNEXTITEM, (WPARAM)-1, LVNI_FOCUSED);

		  if (index != ~0)
		    (*lv_prop->F_dblclick)(hWnd, &lv_prop->items[index], index);}
	   break;


   case WM_CLOSE :
     ShowWindow(hWnd, SW_HIDE);
     PostQuitMessage(NULL);
     break;



   default :
	 return DefWindowProc(hWnd, uMsg, wParam, lParam);}

 return 0;}



LRESULT stack_window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{inputdialog_params iparams;
 SCROLLINFO si;
 sDEFAULT stack_values[8];
 HDC hdc;
 HGDIOBJ gdi_obj;
 int i, j, xpos;
 LRESULT res;
 
 int click_pos;
 sADDR refresh_addr;
 sDEFAULT refresh_data;

 char	buff[64];

 switch (uMsg)
   {case WM_PAINT:
      CallWindowProc(old_stack_wndproc, hWnd, uMsg, wParam, lParam);
	  idb_read(stack_view_addr, stack_values, 32);
      hdc = GetDC(hWnd);
      gdi_obj = SelectObject(hdc, (HGDIOBJ)dialog_font);
	  SetBkColor(hdc, RGB(236, 233, 216));

      for (i = 0; i < 8; i++)
        {if (stack_view_addr + i*4 == context.R_sp) SetTextColor(hdc, RGB(255, 0, 0));
           else SetTextColor(hdc, RGB(0, 0, 0));
	     memset(buff, ' ', sizeof(buff));
	     QXhextostr0A_32(stack_view_addr + i*4, buff, 0);
	     lstrcat(buff, "    ");
	     QXhextostr0A_32(stack_values[i], buff + 12, 0);
	     lstrcat(buff, "    ");
	   
	     if (idb_valid(stack_values[i], 24))
		   {idb_read(stack_values[i], buff+24, 24);
	        for (j = 24; j < 48; j++)
	          {if (buff[j] == 0x00)
	             {for (j; j < 48; j++)
	                buff[j] = ' ';
		            break;}}}
	     else buff[24] = ' ';
	   
	     TextOut(hdc, 8, 20 + i*16, buff, 8 + 4 + 8 + 4 + 24);}

      SelectObject(hdc, gdi_obj);
   
      ReleaseDC(hWnd, hdc);
      break;

 case WM_VSCROLL :
   si.cbSize = sizeof(SCROLLINFO);
   si.fMask = SIF_ALL;
   GetScrollInfo(hWnd, SB_VERT, &si);
   xpos = si.nPos;
 
   switch (wParam&0xFFFF)
     {case SB_LINEUP :
        if (xpos >= 1)
		  {stack_view_addr -= 4;
           xpos--;}
		break;
		
	  case SB_LINEDOWN :
        if (xpos <= (0x10000 / 4) - 1)
		  {stack_view_addr += 4;
		   xpos++;}
	    break;

	  case SB_PAGEUP :
	    if (xpos >= 8)
		  {stack_view_addr -= 32;
		   xpos -= 8;}
		break;

      case SB_PAGEDOWN :
        if (xpos <= (0x10000 / 4) - 8)
		  {stack_view_addr += 32;
		   xpos += 8;}
		break;

	  case SB_THUMBTRACK :
		stack_view_addr = 4*(si.nTrackPos - (0x10000 / 8)) + context.R_sp;
		break;

	  case SB_THUMBPOSITION :
		xpos = si.nTrackPos;
		stack_view_addr = 4*(xpos - (0x10000 / 8)) + context.R_sp;
		break;}
     
    si.cbSize = sizeof(SCROLLINFO);
    si.fMask = SIF_POS; 
    si.nPos = xpos; 
    SetScrollInfo(hWnd, SB_VERT, &si, TRUE); 

	InvalidateRect(hWnd, NULL, true);
	break;


   case WM_RBUTTONDBLCLK :
     refresh_stack(hWnd);
	 break;

   case WM_LBUTTONDBLCLK :
     click_pos = ((lParam>>16) - 20) / 16;
     if (click_pos >= 0 && click_pos <= 7)
       {refresh_addr = 4*click_pos + stack_view_addr;
	    idb_read(refresh_addr, &refresh_data, sizeof(sDEFAULT));

		wsprintf(buff, "Enter new value for address %lX in stack area", refresh_addr);
		iparams.comment = buff;
		iparams.F_in = &f_in_32bit;
		iparams.F_out = &f_out_32bit;
		iparams.item_data = &refresh_data;
		iparams.item_number = NULL;
		if (DialogBoxParam(plugin_instance, MAKEINTRESOURCE(InputDialog),
	                       hWnd, (DLGPROC)&inputdlg_window_proc, (LPARAM)&iparams))
		  {idb_write(refresh_addr, &refresh_data, sizeof(sDEFAULT));
		   InvalidateRect(hWnd, NULL, true);}}
     break;


   default : return DefWindowProc(hWnd, uMsg, wParam, lParam);}
 return 0;}


LRESULT dumpdialog_window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{HANDLE hFile;
 static sDEFAULT operation;
 
 DWORD hFileSize;
 sADDR addr;
 sDEFAULT size;
 void *dump;

 char buff[MAX_PATH], buff1[16], buff2[16];
 DWORD tmp0;
 
 switch (uMsg)
   {case WM_CLOSE :
	  EndDialog(hWnd, NULL);
      return true;
 
    case WM_INITDIALOG :
	  operation = ((dumpdialog_params*)lParam)->operation;
	  SetWindowText(hWnd, ((dumpdialog_params*)lParam)->title);
   
      if (operation == OPERATION_WRITE)
        EnableWindow(GetDlgItem(hWnd, dump_SizeButton), false);
	  else
        EnableWindow(GetDlgItem(hWnd, dump_SizeButton), true);
	  return true;

	case WM_COMMAND :
      if ((wParam&0xFFFF) == dump_BrowseButton && (wParam>>16) == BN_CLICKED)
        {OpenDialogExecute(plugin_instance, hWnd, 
		                   "dump files(*.dmp)\0*.dmp\0All files(*.*)\0*.*\0\0",
						   buff);
		 SetDlgItemText(hWnd, dump_FNEdit, buff);}
	  

      if ((wParam&0xFFFF) == dump_SizeButton && (wParam>>16) == BN_CLICKED)
	    {GetDlgItemText(hWnd, dump_FNEdit, buff, MAX_PATH);
	     hFile = CreateFile(buff, GENERIC_READ, FILE_SHARE_READ,
			                NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		 if (hFile == INVALID_HANDLE_VALUE)
		   {MessageBox(hWnd, "Cannot open file", buff, MB_ICONERROR);
		    return true;}
		 QXhextostr0A_32(GetFileSize(hFile, NULL), buff, 0);
		 CloseHandle(hFile);
		 SetDlgItemText(hWnd, dump_SizeEdit, buff);}

      if ((wParam&0xFFFF) == IDOK && (wParam>>16) == BN_CLICKED)
	    {GetDlgItemText(hWnd, dump_FNEdit, buff, MAX_PATH);
	     GetDlgItemText(hWnd, dump_AddrEdit, buff1, 16);
		 GetDlgItemText(hWnd, dump_SizeEdit, buff2, 16);

         if (!QXstrtohex_32(buff1, (long*)&addr) || !QXstrtohex_32(buff2, (long*)&size))
		   {MessageBox(hWnd, "Invalid address of size", "ERROR!!!", MB_ICONERROR);
		    return true;}

         if (operation == OPERATION_READ)
           {hFile = CreateFile(buff, GENERIC_READ, FILE_SHARE_READ,
			                   NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		    if (hFile == INVALID_HANDLE_VALUE)
		      {MessageBox(hWnd, "Cannot open file", buff, MB_ICONERROR);
		       return true;}

		    hFileSize = GetFileSize(hFile, NULL);
			if (size > hFileSize)
		      {CloseHandle(hFile);
			   MessageBox(hWnd, "Entered size is greater than real file size", "ERROR!!!", MB_ICONERROR);
		       return true;}
            dump = LocalAlloc(LMEM_FIXED, size);
			ReadFile(hFile, dump, size, &tmp0, NULL);
            idb_write(addr, dump, size);
			CloseHandle(hFile);
			EndDialog(hWnd, NULL);
			return true;}

		 else if (operation == OPERATION_WRITE)
		   {hFile = CreateFile(buff, GENERIC_WRITE, FILE_SHARE_READ,
			                   NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		    if (hFile == INVALID_HANDLE_VALUE)
		      {MessageBox(hWnd, "Cannot open file", buff, MB_ICONERROR);
		       return true;}

		    hFileSize = GetFileSize(hFile, NULL);
            dump = LocalAlloc(LMEM_FIXED, size);
            idb_read(addr, dump, size);
			WriteFile(hFile, dump, size, &tmp0, NULL);
			CloseHandle(hFile);
			EndDialog(hWnd, NULL);
			return true;}}

	  return true;}
 return false;}




void InitComboBoxes(HWND hWnd)
{int i;

 for (i = 0; i < MAIN_VIEWS_NUMBER; i++)
   {SendDlgItemMessage(hWnd, ViewCombo1, CB_ADDSTRING, NULL, (LPARAM)main_views[i]);
    SendDlgItemMessage(hWnd, ViewCombo2, CB_ADDSTRING, NULL, (LPARAM)main_views[i]);}
 SendDlgItemMessage(hWnd, ViewCombo1, CB_SETCURSEL, 0, NULL);
 SendDlgItemMessage(hWnd, ViewCombo2, CB_SETCURSEL, 1, NULL);}


void InitListViews(HWND hWnd)
{int i;
 HWND listview_hwnd;
 DWORD exstyle;
	
 for (i = 0; i < sizeof(LVs)/sizeof(listview_content*); i++)
   SendDlgItemMessage(hWnd, LVs[i]->ID, LVM_SETEXTENDEDLISTVIEWSTYLE, ~0, LVS_EX_FULLROWSELECT);}

void InitStackView(HWND hWnd)
{SCROLLINFO si;
 HWND Stack_hwnd;

 Stack_hwnd = GetDlgItem(hWnd, Stack);
 old_stack_wndproc = (WNDPROC)GetWindowLong(Stack_hwnd, GWL_WNDPROC);
 SetWindowLong(Stack_hwnd, GWL_WNDPROC, (LONG)(&stack_window_proc));

 si.cbSize = sizeof(SCROLLINFO);
 si.fMask = SIF_ALL;
 si.nMin = 0;
 si.nMax = 0x10000 / 4 - 1;
 si.nPage = 8;
 si.nPos = 0x10000 / 4 / 2;
 SetScrollInfo(Stack_hwnd, SB_VERT, &si, true);}


void refresh_listview(HWND hWnd, listview_content *LV)
{LV_COLUMN lvc;
 LV_ITEM lvi;
 RECT rect;
 HWND listview_hwnd;
 int i;
 char buff[64];

 listview_hwnd = GetDlgItem(hWnd, LV->ID);

 lvi.mask = LVIF_TEXT;
 lvi.pszText = buff;

 if (LV->old_prop != LV->prop ||
	 LV->prop->count != SendMessage(listview_hwnd, LVM_GETITEMCOUNT, NULL, NULL))
   {SendMessage(listview_hwnd, LVM_DELETEALLITEMS, NULL, NULL);
    SendMessage(listview_hwnd, LVM_DELETECOLUMN, 0, NULL);
    SendMessage(listview_hwnd, LVM_DELETECOLUMN, 0, NULL);
 
    lvc.mask = LVCF_TEXT+LVCF_WIDTH;
    GetClientRect(listview_hwnd, &rect);

    lvc.pszText = LV->prop->column1;
	lvc.cx = (rect.right * LV->prop->c1_width)/100;
	SendMessage(listview_hwnd, LVM_INSERTCOLUMN, 1, (LPARAM)&lvc);

	lvc.pszText = LV->prop->column2;
	lvc.cx = (rect.right * LV->prop->c2_width)/100;
	SendMessage(listview_hwnd, LVM_INSERTCOLUMN, 2, (LPARAM)&lvc);

	lvi.iSubItem = 0;
	for (i = 0; i < LV->prop->count; i++)
      {lvi.iItem = i;
	   LV->prop->items[i].F_descr_out(LV->prop->items->item_data, i, buff);
	   SendMessage(listview_hwnd, LVM_INSERTITEM, NULL, (LPARAM)&lvi);}
   
	LV->old_prop = LV->prop;}

 lvi.iSubItem = 1;
 for (i = 0; i < LV->prop->count; i++)
   {LV->prop->items[i].F_out(LV->prop->items[i].item_data, i, buff);
    SendMessage(listview_hwnd, LVM_SETITEMTEXT, i, (LPARAM)&lvi);}}

void refresh_stack(HWND hWnd)
{SCROLLINFO si;
 HWND stack_hwnd;

 stack_hwnd = GetDlgItem(hWnd, Stack);
 stack_view_addr = context.R_sp - 4*2;
 
 si.cbSize = sizeof(SCROLLINFO);
 si.fMask = SIF_POS;
 si.nPos = 0x10000 / 8 - 2;
 SetScrollInfo(stack_hwnd, SB_VERT, &si, true);

 InvalidateRect(stack_hwnd, NULL, true);}


void refresh_status(HWND hWnd)
{HWND button_hwnd;

 button_hwnd = GetDlgItem(hWnd, StatusButton);

 if (emulator_enable)
   {EnableWindow(button_hwnd, true);
    SetWindowText(button_hwnd, "Emulator status: RUN, Press this button for stop emulate");}
 else
   {EnableWindow(button_hwnd, false);
    SetWindowText(button_hwnd, "Emulator status: WAIT");}}


void refresh_all(HWND hWnd)
{int i;

 if (context.breakpoints_number < 32)
   BP_listview.count = context.breakpoints_number + 1;
 else BP_listview.count = 32;

 for (i = 0; i < sizeof(LVs)/sizeof(listview_content*); i++)
   refresh_listview(hWnd, LVs[i]);

 refresh_stack(hWnd);
 refresh_status(hWnd);}





bool inputdlg_window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{bool res;
 char buffer[128];
 static inputdialog_params *params;

 switch(uMsg)
   {case WM_COMMAND :
      if ((wParam&0xFFFF) == IDOK)
	    {GetDlgItemText(hWnd, input_DataEdit, buffer, 127);
	     if (!(*params->F_in)(params->item_data, params->item_number, buffer))
           MessageBox(hWnd, "You have entered invalid data", "Error", MB_ICONERROR);
		 else
		   EndDialog(hWnd, true);}
	  
	  else if ((wParam&0xFFFF) == IDCANCEL)
	    EndDialog(hWnd, false);
	  break;
 
    case WM_INITDIALOG :
      params = (inputdialog_params*)lParam;
	  (*params->F_out)(params->item_data, params->item_number, buffer);

	  SetDlgItemText(hWnd, input_CommentView, params->comment);
      SetDlgItemText(hWnd, input_DataEdit, buffer);
	  SendDlgItemMessage(hWnd, input_DataEdit, EM_SETSEL, 0, ~0);
	  SetFocus(GetDlgItem(hWnd, input_DataEdit));
	  return false;

    case WM_CLOSE :
	  EndDialog(hWnd, false);
	  break;}

 return false;}


void f_out_16bit(void *data, sDEFAULT number, char *out)
{QXhextostr0A_16(*(short*)data, out, 0);}

bool f_in_16bit(void *out, sDEFAULT number, char *data)
{int i;

 for (i = 0; i < SEG_REGISTERS_NUMBER; i++)
   if (!lstrcmpi(data, reg_segment_ctrl[i]))
     {*(sWORD*)out = *seg_ctrl_register_values[i];   
      return true;}
 return QXstrtohex_16(data, (short*)out);}


void f_out_32bit(void *data, sDEFAULT number, char *out)
{QXhextostr0A_32(*(long*)data, out, 0);}


bool f_in_32bit(void *out, sDEFAULT number, char *data)
{int i;

 for (i = 0; i < GENERAL_REGISTERS_NUMBER; i++)
   if (!lstrcmpi(data, reg_general[i]))
     {*(sDWORD*)out = *general_register_values[i];   
      return true;}
 if (!lstrcmpi(data, "eip"))
   {*(sDWORD*)out = context.R_ip;   
    return true;}
 return QXstrtohex_32(data, (long*)out);}


void f_out_64bit(void *data, sDEFAULT number, char *out)
{QXhextostr0A_64(*(long*)data, out, 0);}

bool f_in_64bit(void *out, sDEFAULT number, char *data)
{int i;

 for (i = 0; i < MMX_REGISTERS_NUMBER; i++)
   if (!lstrcmpi(data, reg_mmx[i]))
     {*(sQWORD*)out = *mmx_register_values[i];   
      return true;}
 return QXstrtohex_64(data, (__int64*)out);}

void f_out_fpu(void *data, sDEFAULT number, char *out)
{_gcvt(*(long double*)data, 18, out);}

bool f_in_fpu(void *out, sDEFAULT number, char *data)
{int i;

 for (i = 0; i < FPU_REGISTERS_NUMBER; i++)
   if (!lstrcmpi(data, reg_fpu[i]))
     {*(long double*)out = *fpu_register_values[i];   
      return true;}
	
	
 *(long double*)out = atof(data);
 return true;}

#define bp ((breakpoint*)out)
bool f_in_bp(void *out, sDEFAULT number, char *data)
{char buff1[32];
 char *ptr;
 
 sDEFAULT type = 0, addr = 0, size = 0;
 int i;


 ptr = QXstrcpytn(buff1, data, ' ', 4);
 if (*buff1 == '\0')
   {bp->type = ~0;
    return true;}

 else if (!lstrcmpi(buff1, "bpx"))
   {QXstrcpytn(buff1, ptr, ' ', 10);
    
    if (f_in_32bit(&addr, 0, buff1))
	  {bp->type = BREAKPOINT_EXECUTE;
	   bp->size = 1;
	   bp->addr = addr;
	   return true;}}

 else if (!lstrcmpi(buff1, "bpm") ||
	      !lstrcmpi(buff1, "bpmb") ||
		  !lstrcmpi(buff1, "bpmw") ||
		  !lstrcmpi(buff1, "bpmd"))
   {if (buff1[3] == 'w')
      size = 2;
    else if (buff1[3] == 'd')
	  size = 4;
	else
	  size = 1;

	ptr = QXstrcpytn(buff1, ptr, ' ', 10);
	if (f_in_32bit(&addr, 0, buff1))
	  {QXstrcpytn(buff1, ptr, ' ', 2);
	   
	   if (!lstrcmpi(buff1, "r"))
	     type = BREAKPOINT_READ;
	   else if (!lstrcmpi(buff1, "w"))
	     type = BREAKPOINT_WRITE;
	   else if (!lstrcmpi(buff1, "rw") || *buff1 == 0)
	     type = BREAKPOINT_READ|BREAKPOINT_WRITE;
	   else if (!lstrcmpi(buff1, "x"))
	     type = BREAKPOINT_EXECUTE;
	   else
	     return false;

       bp->type = type;
	   bp->size = size;
	   bp->addr = addr;
       return true;}}

 else if (!lstrcmpi(buff1, "bpr") || 
	      !lstrcmpi(buff1, "!bpr"))
   {ptr = QXstrcpytn(buff1, ptr, ' ', 3);
   
    for (i = 0; i < 3; i++)
	  {switch (buff1[i])
	    {case 'r' :
	       if (!(type&BREAKPOINT_READ))
		     {type |= BREAKPOINT_READ;
		      break;}
		   else
		     return false;

		 case 'w' :
	       if (!(type&BREAKPOINT_WRITE))
		     {type |= BREAKPOINT_WRITE;
		      break;}
		   else
		     return false;

		 case 'x' :
	       if (!(type&BREAKPOINT_EXECUTE))
		     {type |= BREAKPOINT_EXECUTE;
		      break;}
		   else
		     return false;

		 case '\0' :
		   break;

		 default :
		   return false;}}
    
    if (!type) return false;

    ptr = QXstrcpytn(buff1, ptr, ' ', 10);
    if (f_in_32bit(&addr, 0, buff1))
	  {ptr = QXstrcpytn(buff1, ptr, ' ', 10);
	   if (f_in_32bit(&size, 0, buff1))
	     {bp->type = type;
	      bp->size = size;
		  bp->addr = addr;
		  return true;}}}
 else
   return false;
 return false;}
#undef bp

#define bp ((breakpoint*)data)
void f_out_bp(void *data, sDEFAULT number, char *out)
{if (bp->type == ~0)
   {lstrcpy(out, "... no breakpoint ...");}
 
 else if (bp->type&BREAKPOINT_EXECUTE && bp->size == 1)
   {lstrcpy(out, "bpx ");
    QXhextostr0A_32(bp->addr, out+4, 0);}
 else if ((bp->size == 1 || bp->size == 2 || bp->size == 4) && 
	      (!(bp->type&BREAKPOINT_EXECUTE) || bp->type == BREAKPOINT_EXECUTE))
   {lstrcpy(out, "bpm");
    if (bp->size == 1) lstrcat(out, "b ");
	else if (bp->size == 2) lstrcat(out, "w ");
	else lstrcat(out, "d ");

	QXhextostr0A_32(bp->addr, out + 5, 0);
	lstrcat(out, " ");

	if (bp->type&BREAKPOINT_READ) lstrcat(out, "r");
	if (bp->type&BREAKPOINT_WRITE) lstrcat(out, "w");
    if (bp->type&BREAKPOINT_EXECUTE) lstrcat(out, "x");}

 else
   {lstrcpy(out, "bpr ");
	
    if (bp->type&BREAKPOINT_READ) lstrcat(out, "r");
	if (bp->type&BREAKPOINT_WRITE) lstrcat(out, "w");
	if (bp->type&BREAKPOINT_EXECUTE) lstrcat(out, "x");
	lstrcat(out, " ");

	QXhextostr0A_32(bp->addr, out + lstrlen(out), 0);
	lstrcat(out, " ");

	QXhextostr0A_32(bp->size, out + lstrlen(out), 0);}}
#undef bp


void f_out_flags(void *data, sDEFAULT number, char *out)
{static sDEFAULT flags_mask[] = {FLAG_C, FLAG_P, FLAG_A, FLAG_Z, FLAG_S,
								 FLAG_T, FLAG_I, FLAG_D, FLAG_O};
 if (*(sDEFAULT*)data&flags_mask[number])
   *out = '1';
 else *out = '0';
 *(out + 1) = 0;}


void f_GP_descr(void *data, sDEFAULT number, char *out)
{lstrcpy(out, reg_general[number]);}

void f_MMX_descr(void *data, sDEFAULT number, char *out)
{lstrcpy(out, reg_mmx[number]);}

void f_FPU_descr(void *data, sDEFAULT number, char *out)
{lstrcpy(out, reg_fpu[number]);}

void f_CR_TR_descr(void *data, sDEFAULT number, char *out)
{lstrcpy(out, reg_ctrl[number]);}
void f_DR_descr(void *data, sDEFAULT number, char *out)
{lstrcpy(out, reg_dbg[number]);}

void f_SEG_descr(void *data, sDEFAULT number, char *out)
{lstrcpy(out, reg_segment_ctrl[number]);}

void f_FLAGS_descr(void *data, sDEFAULT number, char *out)
{lstrcpy(out, flags[number]);}


void f_bp_descr(void *data, sDEFAULT number, char *out)
{wsprintf(out, "%d", number+1);}



void f_dblclick_regview(HWND hWnd, item_properties *item, int item_number)
{char buff1[8], buff2[64];
 inputdialog_params iparams;

 (*item->F_descr_out)(item->item_data, item_number, buff1);
 wsprintf(buff2, "Enter new value for register %s:", buff1);

 iparams.comment = buff2;
 iparams.F_in = item->F_in;
 iparams.F_out = item->F_out;
 iparams.item_data = item->item_data;
 iparams.item_number = item_number;
 if (DialogBoxParam(plugin_instance, MAKEINTRESOURCE(InputDialog),
	                hWnd, (DLGPROC)&inputdlg_window_proc, (LPARAM)&iparams))
   refresh_all(hWnd);}


void f_dblclick_bp(HWND hWnd, item_properties *item, int item_number)
{char buff[64];
 inputdialog_params iparams;

 wsprintf(buff, "Enter new value for breakpoint %d:", item_number + 1);

 iparams.comment = buff;
 iparams.F_in = item->F_in;
 iparams.F_out = item->F_out;
 iparams.item_data = item->item_data;
 iparams.item_number = item_number;
 if (DialogBoxParam(plugin_instance, MAKEINTRESOURCE(InputDialog),
	                hWnd, (DLGPROC)&inputdlg_window_proc, (LPARAM)&iparams))
   {if (((breakpoint*)item->item_data)->type == ~0)
      {if (item_number < (int)context.breakpoints_number - 1)
         QXmemcpy(&context.breakpoints[item_number],
		          &context.breakpoints[item_number + 1],
				  sizeof(breakpoint)*(context.breakpoints_number - item_number - 1));

       if (item_number < context.breakpoints_number)
	     {context.breakpoints_number--;
	      BP_listview.count--;

		  context.breakpoints[context.breakpoints_number].type = ~0;}}

   else if (item_number == context.breakpoints_number &&
		     context.breakpoints_number < 32)
	  {context.breakpoints_number++;
       context.breakpoints[context.breakpoints_number].type = ~0;
	   
	   BP_listview.count++;}

   refresh_all(hWnd);}}


void f_dblclick_flags(HWND hWnd, item_properties *item, int item_number)
{context.R_flags ^= flags_mask[item_number];
 refresh_all(hWnd);}





bool OpenDialogExecute(HINSTANCE hInstance, HWND owner, char *filter, char *filename)
{OPENFILENAME ofn;

 QXZeroMemory(&ofn, sizeof(OPENFILENAME));
 *filename = 0;

 ofn.lStructSize = sizeof(OPENFILENAME);
 ofn.hInstance = hInstance;
 ofn.hwndOwner = owner;
 ofn.lpstrFilter = filter;
 ofn.lpstrFile = filename;
 ofn.nMaxFile = MAX_PATH;
 ofn.Flags = OFN_FILEMUSTEXIST + OFN_PATHMUSTEXIST;
 ofn.lpstrTitle = "Select and open file";
 return GetOpenFileName(&ofn);}




bool SaveDialogExecute(HINSTANCE hInstance, HWND owner, char *filter, char *filename)
{OPENFILENAME ofn;

 QXZeroMemory(&ofn, sizeof(OPENFILENAME));
 *filename = 0;

 ofn.lStructSize = sizeof(OPENFILENAME);
 ofn.hInstance = hInstance;
 ofn.hwndOwner = owner;
 ofn.lpstrFilter = filter;
 ofn.lpstrFile = filename;
 ofn.nMaxFile = MAX_PATH;
 ofn.Flags = OFN_PATHMUSTEXIST;
 ofn.lpstrTitle = "Select and save file";
 return GetSaveFileName(&ofn);}








int QXhextostr0A_16(short num, char *buff, unsigned long sign)
{char *ptr, sym, i;

 if (!buff) ptr = (char*)LocalAlloc(LPTR, 5);
   else ptr = buff;
 if (sign && num&0x8000)
   {*ptr = '-';
    ptr++;
	num = -num;}
 i = 0;
 while (i < 4)
   {sym = (num >> 12)&0xF;
    if (sym > 9) sym += ('A' - '9' - 1);
	*ptr = (sym + '0');
	ptr++;
    num <<= 4;
	i++;}
 *ptr = 0;
 return 4;}


int QXhextostr0A_32(long num, char *buff, unsigned long sign)
{char *ptr, sym, i;

 if (!buff) ptr = (char*)LocalAlloc(LPTR, 9);
   else ptr = buff;
 if (sign && num&0x80000000)
   {*ptr = '-';
    ptr++;
	num = -num;}
 i = 0;
 while (i < 8)
   {sym = (num >> 28)&0xF;
    if (sym > 9) sym += ('A' - '9' - 1);
	*ptr = (sym + '0');
	ptr++;
    num <<= 4;
	i++;}
 *ptr = 0;
 return 8;}



int QXhextostr0A_64(__int64 num, char *buff, unsigned long sign)
{char *ptr, sym, i;

 if (!buff) ptr = (char*)LocalAlloc(LPTR, 17);
   else ptr = buff;
 if (sign && num&0x8000000000000000)
   {*ptr = '-';
    ptr++;
	num = -num;}
 i = 0;
 while (i < 16)
   {sym = (num >> 60)&0xF;
    if (sym > 9) sym += ('A' - '9' - 1);
	*ptr = (sym + '0');
	ptr++;
    num <<= 4;
	i++;}
 *ptr = 0;
 return 4;}




bool QXstrtohex_16(char *hex, short *out)
{short res;
 char *bptr, *eptr, sym, exp;

 bptr = hex;
 if (*(WORD*)bptr == 0x7830) bptr += 2;

 eptr = hex + lstrlen(hex) - 1;
 if (*eptr == 'h') eptr--;

 res = 0;
 exp = 0;
 do {sym = *eptr;
 
     if (sym >= '0' && sym <= '9') sym -= '0';
	 else if (sym >= 'A' && sym <= 'F') sym -= ('A' - 10);
	      else if (sym >= 'a' && sym <= 'f') sym -= ('a' - 10);
	           else return false;
	 
	 res |= ((short)sym << exp);
	 exp += 4;}
 while (--eptr >= bptr && exp < 16);

 *out = res;
 return true;}

bool QXstrtohex_32(char *hex, long *out)
{long res;
 char *bptr, *eptr, sym, exp;

 bptr = hex;
 if (*(WORD*)bptr == 0x7830) bptr += 2;

 eptr = hex + lstrlen(hex) - 1;
 if (*eptr == 'h') eptr--;

 res = 0;
 exp = 0;
 do {sym = *eptr;
 
     if (sym >= '0' && sym <= '9') sym -= '0';
	 else if (sym >= 'A' && sym <= 'F') sym -= ('A' - 10);
	      else if (sym >= 'a' && sym <= 'f') sym -= ('a' - 10);
	           else return false;
	 
	 res |= ((long)sym << exp);
	 exp += 4;}
 while (--eptr >= bptr && exp < 32);

 *out = res;
 return true;}



bool QXstrtohex_64(char *hex, __int64 *out)
{__int64 res;
 char *bptr, *eptr, sym, exp;

 bptr = hex;
 if (*(WORD*)bptr == 0x7830) bptr += 2;

 eptr = hex + lstrlen(hex) - 1;
 if (*eptr == 'h') eptr--;

 res = 0;
 exp = 0;
 do {sym = *eptr;
 
     if (sym >= '0' && sym <= '9') sym -= '0';
	 else if (sym >= 'A' && sym <= 'F') sym -= ('A' - 10);
	      else if (sym >= 'a' && sym <= 'f') sym -= ('a' - 10);
	           else return false;
	 
	 res |= ((__int64)sym << exp);
	 exp += 4;}
 while (--eptr >= bptr && exp < 64);

 *out = res;
 return true;}



char* QXstrcpyt(char *buff, char *str, char delimiter)
{char *sptr = str,
      *dptr = buff;

 while (*sptr == delimiter) sptr++;
 while (*sptr != delimiter && *sptr != '\0') *dptr++ = *sptr++;
 
 *dptr = '\0';
 return sptr;}


char* QXstrcpytn(char *buff, char *str, char delimiter, int len)
{char *sptr = str,
      *dptr = buff;
 int i;

 i = 0;
 while (*sptr == delimiter) sptr++;
 while (*sptr != delimiter && *sptr != '\0' && i++ < len) *dptr++ = *sptr++;
 
 *dptr = '\0';
 return sptr;}


