#define __NT__
#define __IDP__

#include "main.h"
#include "emul.h"



#define EMUL_STATUS_BREAKPOINT				0x00010000


sADDR				stack_segment_base = 0;
sADDR				peb_segment_base = 0;

	//========================= Emulator context structure ========================= //
x86_emul_context	context;


//==================== Selector associates for PEB and other segments ====================//
selector_associate		selectors[3] = {
{0x1B, 0x00000000},
{0x23, 0x00000000},
{0x38, 0x7FFDF000}};

//==================== Pointers to general registers in context structure ====================//
sDEFAULT *general_register_values[] = {
&context.R_ax, &context.R_cx, &context.R_dx, &context.R_bx, 
&context.R_sp, &context.R_bp, &context.R_si, &context.R_di};


//==================== Pointers to FPU registers ====================//
long double *fpu_register_values[] = {
&context.R_st0, &context.R_st1, &context.R_st2, &context.R_st3, 
&context.R_st4, &context.R_st5, &context.R_st6, &context.R_st7};

//==================== Pointers to MMX registers ====================//
unsigned __int64 *mmx_register_values[] = {
&context.R_MM0, &context.R_MM1, &context.R_MM2, &context.R_MM3, 
&context.R_MM4, &context.R_MM5, &context.R_MM6, &context.R_MM7};


//==================== Pointers to segment registers ======================//
unsigned __int16 *segment_register_values[] = {
&context.R_es, &context.R_cs, &context.R_ss, &context.R_ds, &context.R_fs, &context.R_gs};


//==================== Pointers to control registers ======================//
sDEFAULT *control_register_values[] = {
&context.R_cr0, &context.R_cr2, &context.R_cr3, &context.R_cr4};

//==================== Pointers to debug registers ======================//
sDEFAULT *debug_register_values[] = {
&context.R_dr0, &context.R_dr1, &context.R_dr2, &context.R_dr3, &context.R_dr6, &context.R_dr7};


	//==================== Operands properties ====================//
operand_properties	op_prop[MAX_OPERANDS_NUMBER];


	//==================== Operands cache ====================//
fixed_cache			op_cache[MAX_OPERANDS_NUMBER];


	//==================== Additional Instruction Params ====================//
additional_instr_params		instr_params;


	//==================== Emulator enable flag ====================//
int					emulator_enable = 0;

	//==================== Instruction tables ====================//

/* This variables initialize in reset_configuration function,
   memory allocate by LocalAlloc function */
instr_properties	*indexed_IT;


extern HWND					hDlg;

/* sizes table */
sDEFAULT sizes_table[] = {
1,		//dt_byte
2,		//dt_word
4,		//dt_dword
4,		//dt_float
8,		//dt_double
10,		//dt_tbyte
0,		//dt_packreal
8,		//dt_qword
16,		//dt_byte16
0,		//dt_code
0,		//dt_void
6,		//dt_fword
0,		//dt_bitfild
0,		//dt_string
0		//dt_unicode
};

sDEFAULT sizes[] = {
0,		//dt_byte
1,		//dt_word
2,		//dt_dword
2,		//dt_float
3,		//dt_double
0x80,	//dt_tbyte
0xFF,	//dt_packreal
3,		//dt_qword
4,		//dt_byte16
0xFF,	//dt_code
0xFF,	//dt_void
0x81,	//dt_fword
0xFF,	//dt_bitfild
0xFF,	//dt_string
0xFF	//dt_unicode
};

sDEFAULT ctrl_conv[] = {0, 0xFF, 1, 2, 3, 0xFF, 0xFF, 0xFF, 4};
sDEFAULT dbg_conv[] = {0, 1, 2, 3, 4, 5, 4, 5};

	//Functions for retrieve operand value
bool	get_operand_value(op_t *operand, operand_properties *ptr);
int		get_operand_value_ch(op_t *operand, operand_properties *ptr, sBYTE mask);

	//Functions for calculate operand address
bool	get_operand_address(op_t *operand, operand_properties *ptr, fixed_cache *cache, sBYTE mask);
int		get_operand_address_ch(op_t *operand, operand_properties *ptr, fixed_cache *cache, sBYTE mask);

	//And this function calculates size of operand
bool	get_operand_size(op_t *operand, operand_properties *ptr);

	//Address calculator
sDEFAULT	calculate_addr(op_t *operand);

	//Function for save cached operands to base
void	flush_operands();


	//Function for get base address for memory operand
sADDR	get_memref_base(op_t *operand);

	//Testing memory area
bool	MemoryExists(sADDR addr, sDEFAULT size);

	//Prepare emulation
void	prepare_emul();



bool	reset_configuration()
{QXZeroMemory(&context, sizeof(context));
 
 context.R_ip = get_screen_ea();
 context.R_flags = FLAG_I;

 context.R_cs = 0x1B;
 context.R_ds = context.R_ss = context.R_es = 0x23;
 context.R_fs = 0x38;

 context.breakpoints[0].type = ~0;

 create_stack_segment();
 context.R_sp = stack_segment_base + 0xFFC4;
 
 create_peb_segment();
 return true;}


bool	trace_into(sADDR addr)
{bool res = emulate_instruction(addr);
 if (res && !isCode(getFlags(context.R_ip)) && !isData(getFlags(context.R_ip)))
   {do_unknown_range(context.R_ip, 5, false);
    ua_code(context.R_ip);}
 refresh_all(hDlg);
 jumpto(context.R_ip);
 return res;}






int		run_to_cursor(sADDR addr)
{int success_emulate = 1;
	
 while ((emulator_enable == 1) && (context.R_ip != addr) && (success_emulate == 1))
   success_emulate = emulate_instruction(context.R_ip);

 if (success_emulate && !isCode(getFlags(context.R_ip)) && !isData(getFlags(context.R_ip)))
   {do_unknown_range(context.R_ip, 5, false);
    ua_code(context.R_ip);}

 emulator_enable = 0;
 refresh_all(hDlg);
 jumpto(context.R_ip);
 return 0;}



int		step_over(int unk0)
{int success_emulate = 1;

 
 ua_ana0(context.R_ip);
 if (cmd.itype == NN_call)
   {context.sys_breakpoint_addr = context.R_ip + cmd.size;

    while (emulator_enable && success_emulate)
      success_emulate = emulate_instruction_ch(context.R_ip);}

 else if (cmd.auxpref&aux_rep || cmd.auxpref&aux_repne)
   {context.sys_breakpoint_addr = context.R_ip + cmd.size;

    while (emulator_enable && success_emulate && context.R_ip != context.sys_breakpoint_addr)
      success_emulate = emulate_instruction(context.R_ip);}

 else emulate_instruction(context.R_ip);   

 emulator_enable = 0;
 refresh_all(hDlg);
 jumpto(context.R_ip);
 return 0;}


int		e_run(int unk0)
{int success_emulate = 1;
	
 while ((emulator_enable == 1) && (success_emulate == 1))
   success_emulate = emulate_instruction_ch(context.R_ip);

 if (success_emulate && !isCode(getFlags(context.R_ip)) && !isData(getFlags(context.R_ip)))
   {do_unknown_range(context.R_ip, 5, false);
    ua_code(context.R_ip);}

 emulator_enable = 0;
 refresh_all(hDlg);
 jumpto(context.R_ip);
 return 0;}




int		execute_till_return(int unk0)
{int success_emulate = 1;
	
 cmd.itype == NN_null;

 while (emulator_enable && success_emulate && cmd.itype != NN_retn)
   success_emulate = emulate_instruction_ch(context.R_ip);

 if (success_emulate && !isCode(getFlags(context.R_ip)) && !isData(getFlags(context.R_ip)))
   {do_unknown_range(context.R_ip, 5, false);
    ua_code(context.R_ip);}

 emulator_enable = 0;
 refresh_all(hDlg);
 jumpto(context.R_ip);
 return 0;}



bool	emulate_instruction(sADDR addr)
{bool result;
 sBYTE mask;
 instr_handler handler;
 int i;

 if (!ua_ana0(addr))
   {msg("address: %lX, cannot analyze code\n", addr);
	return false;}


 if (indexed_IT[cmd.itype].handler == 0)
   {msg("address: %lX, emulator does not support this instruction\n", addr);
    return false;}

 if (cmd.itype == NN_sidt)
   cmd.Op1.dtyp = dt_fword;

 for (i = 0; i < MAX_OPERANDS_NUMBER; i++)
   {mask = indexed_IT[cmd.itype].mask[i];
    if (!mask) break;

	if (mask&OPERAND_VALUE)
    if (!get_operand_value(&cmd.Operands[i], &op_prop[i]))
      {msg("address: %lX, cannot retrieve operand %d\n", addr, i);
       return false;}

    if (mask&OPERAND_address)
    if (!get_operand_address(&cmd.Operands[i], &op_prop[i], &op_cache[i], mask))
      {msg("address: %lX, cannot retrieve address of operand %d\n", addr, i);
       return false;}

    if (!get_operand_size(&cmd.Operands[i], &op_prop[i]))
      {msg("address: %lX, cannot retrieve size of operand %d\n", addr, i);
       return false;}}

 handler = indexed_IT[cmd.itype].handler;

	//push/pop segment registers fix
 if (cmd.itype == NN_push && cmd.Op1.reg >= R_es && cmd.Op1.reg <= R_gs)
   context.R_sp -= 2;

 if (cmd.itype == NN_leavew)
   *(sWORD*)&context.R_sp = *(sWORD*)&context.R_bp;

 if (cmd.itype == NN_leave || cmd.itype == NN_leaved)
   *(sDWORD*)&context.R_sp = *(sDWORD*)&context.R_bp;

	//imul instruction fix
 if (cmd.itype == NN_imul)
   {if (!(cmd.Op1.flags&OF_SHOW))
	  op_prop[0].addr = 0;
    else if (cmd.Op3.type == o_void)
      *(sDEFAULT*)op_prop[2].data = *general_register_values[cmd.Op1.reg];}
	
	//Fill additional instruction parameters structure
 instr_params.instr_size = cmd.size;

 if (indexed_IT[cmd.itype].stack_use)
   get_many_bytes(context.R_sp, instr_params.stack_buffer, indexed_IT[cmd.itype].stack_use);
 if (cmd.auxpref&aux_rep)
   instr_params.prefix = 1;
 else if (cmd.auxpref&aux_repne)
   instr_params.prefix = 2;
 else instr_params.prefix = 0;

	//Execute emulator and adjust instruction pointer, if need
 if (!(*handler)()) context.R_ip += cmd.size;

	//push/pop segment registers fix
 if (cmd.itype == NN_pop && cmd.Op1.reg >= R_es && cmd.Op1.reg <= R_gs)
   context.R_sp += 2;

 flush_operands();
 return true;}


int		emulate_instruction_ch(sADDR addr)
{bool result;
 sBYTE mask;
 instr_handler handler;
 int i, status;

  //Testing execute

 if (context.sys_breakpoint_addr == addr)
   {context.sys_breakpoint_addr = 0;
	return false;}

 for (i = 0; i < context.breakpoints_number; i++)
   {if (context.breakpoints[i].type&BREAKPOINT_EXECUTE && context.breakpoints[i].addr == addr)
	  {msg("breakpoint %d accepted\n", i + 1);
       return false;}}

 if (!ua_ana0(addr))
   {msg("address: %lX, cannot analyze code\n", addr);
	return false;}

 if (indexed_IT[cmd.itype].handler == 0)
   {msg("address: %lX, emulator does not support this instruction\n", addr);
    return false;}


 if (cmd.itype == NN_sidt)
   cmd.Op1.dtyp = dt_fword;

 for (i = 0; i < MAX_OPERANDS_NUMBER; i++)
   {mask = indexed_IT[cmd.itype].mask[i];
    if (!mask) break;

    if (mask&OPERAND_VALUE)
	  {status = get_operand_value_ch(&cmd.Operands[i], &op_prop[i], mask);
	   if (status == 0)
         {msg("address: %lX, cannot retrieve operand %d\n", addr, i);
          return false;}

	   else if (status&EMUL_STATUS_BREAKPOINT)
         {msg("address: %lX, breakpoint %d accepted\n", addr, (status^EMUL_STATUS_BREAKPOINT) + 1);
          return false;}}
	

    if (mask&OPERAND_address)
	  {status = get_operand_address_ch(&cmd.Operands[i], &op_prop[i], &op_cache[i], mask);
	   if (status == 0)
         {msg("address: %lX, cannot retrieve address of operand %d\n", addr, i);
          return false;}

	   else if (status&EMUL_STATUS_BREAKPOINT)
         {msg("address: %lX, breakpoint %d accepted\n", addr, (status^EMUL_STATUS_BREAKPOINT) + 1);
          return false;}}
	

    if (!get_operand_size(&cmd.Operands[i], &op_prop[i]))
      {msg("address: %lX, cannot retrieve size of operand %d\n", addr, i);
       return false;}}

 handler = indexed_IT[cmd.itype].handler;

	//push/pop segment registers fix
 if (cmd.itype == NN_push && cmd.Op1.reg >= R_es && cmd.Op1.reg <= R_gs)
   context.R_sp -= 2;

	//imul instruction fix
 if (cmd.itype == NN_imul)
   {if (!(cmd.Op1.flags&OF_SHOW))
	  op_prop[0].addr = 0;
    else if (cmd.Op3.type == o_void)
      *(sDEFAULT*)op_prop[2].data = *general_register_values[cmd.Op1.reg];}
	
	//Fill additional instruction parameters structure
 instr_params.instr_size = cmd.size;

 if (indexed_IT[cmd.itype].stack_use)
   get_many_bytes(context.R_sp, instr_params.stack_buffer, indexed_IT[cmd.itype].stack_use);
 if (cmd.auxpref&aux_rep)
   instr_params.prefix = 1;
 else if (cmd.auxpref&aux_repne)
   instr_params.prefix = 2;
 else instr_params.prefix = 0;


	//Execute emulator and adjust instruction pointer, if need

 if (!(*handler)()) context.R_ip += cmd.size;

	//push/pop segment registers fix
 if (cmd.itype == NN_pop && cmd.Op1.reg >= R_es && cmd.Op1.reg <= R_gs)
   context.R_sp += 2;

 flush_operands();
 return true;}




bool	get_operand_value(op_t *operand, operand_properties *ptr)
{sADDR base, addr;
 sDEFAULT size;

 switch (operand->type)
   {case o_reg :						//Operand is register
      if (operand->reg <= R_di)								//general registers, eax - edi
        {*(sDEFAULT*)ptr->data = *general_register_values[operand->reg];
		 return true;}

	  if (operand->reg >= R_al && operand->reg <= R_bl)		//general regs al - bl
	    {*(sBYTE*)ptr->data = *general_register_values[operand->reg - R_al] & 0xFF;
		 return true;}

	  if (operand->reg >= R_ah && operand->reg <= R_bh)		//general regs ah - bh
	    {*(sBYTE*)ptr->data = (*general_register_values[operand->reg - R_ah] >> 8) & 0xFF;
		 return true;}

	  if (operand->reg >= R_es && operand->reg <= R_gs)		//segment regs
	    {*(sWORD*)ptr->data = *(sBYTE*)segment_register_values[operand->reg - R_es];
	     return true;}
       return false;										//unknown register

    case o_mem    :
    case o_phrase :						//Operand is memory address
    case o_displ  :
	   base = get_memref_base(operand);
	   if (base == sADDR(-1)) return false;

       size = sizes_table[operand->dtyp];
	   if (!size) return false;
        
       addr = calculate_addr(operand) + base;
	   if (!MemoryExists(addr, size)) return false;

	   get_many_bytes(addr, ptr->data, size);
	   return true;	

    case o_imm :
	 *(sDEFAULT*)ptr->data = operand->value;
	 return true;

	case o_near :
	 *(sDEFAULT*)ptr->data = operand->addr;
	 return true;

	case o_far :
      return false;

	case o_void :
	  *(sDEFAULT*)ptr->data = 0;
	  return true;

	case o_fpreg :				//FPU
	  *(long double*)ptr->data = *fpu_register_values[operand->reg];
	  return true;
 
    case o_mmxreg :				//MMX
	  *(sQWORD*)ptr->data = *mmx_register_values[operand->reg];
	  return true;

    case o_crreg :				//Control
	  if (ctrl_conv[operand->reg] != 0xFF)
	    {*(sDEFAULT*)ptr->data = *control_register_values[ctrl_conv[operand->reg]];
	     return true;}
	  else return false;

    case o_dbreg :				//Debug
	  *(sDEFAULT*)ptr->data = *debug_register_values[dbg_conv[operand->reg]];
	  return true;

   }
 return false;}



int	get_operand_value_ch(op_t *operand, operand_properties *ptr, sBYTE mask)
{sADDR base, addr, bp_addr, bp_limit;
 sDEFAULT size;
 int i;

 switch (operand->type)
   {case o_reg :
      if (operand->reg <= R_di)			//general registers, eax - edi
        {*(sDEFAULT*)ptr->data = *general_register_values[operand->reg];
		 return true;}

	  if (operand->reg >= R_al && operand->reg <= R_bl)		//general regs al - bl
	    {*(sBYTE*)ptr->data = *general_register_values[operand->reg - R_al] & 0xFF;
		 return true;}

	  if (operand->reg >= R_ah && operand->reg <= R_bh)		//general regs ah - bh
	    {*(sBYTE*)ptr->data = (*general_register_values[operand->reg - R_ah] >> 8) & 0xFF;
		 return true;}

	  if (operand->reg >= R_es && operand->reg <= R_gs)		//segment regs
	    {*(sWORD*)ptr->data = *(sBYTE*)segment_register_values[operand->reg - R_es];
	     return true;}
       return false;										//unknown register

    case o_mem :
    case o_phrase :
    case o_displ :
	   base = get_memref_base(operand);
	   if (base == sADDR(-1)) return false;

       size = sizes_table[operand->dtyp];
	   if (!size) return false;
        
       addr = calculate_addr(operand) + base;
	   if (!MemoryExists(addr, size)) return false;

       if (mask&OPERAND_READ)
	   for (i = 0; i < context.breakpoints_number; i++)
         {bp_addr = context.breakpoints[i].addr;
	      bp_limit = bp_addr + context.breakpoints[i].size;
	     
		  if (addr >= bp_addr && addr < bp_limit && context.breakpoints[i].type&BREAKPOINT_READ)
		    return EMUL_STATUS_BREAKPOINT | i;}

	   get_many_bytes(addr, ptr->data, size);
	   return true;	

    case o_imm :
	 *(sDEFAULT*)ptr->data = operand->value;
	 return true;

	case o_near :
	 *(sDEFAULT*)ptr->data = operand->addr;
	 return true;

	case o_far :
      return false;

	case o_void :
	  *(sDEFAULT*)ptr->data = 0;
	  return true;
	  
  	case o_fpreg :				//FPU
	  *(long double*)ptr->data = *fpu_register_values[operand->reg];
	  return true;
 
    case o_mmxreg :				//MMX
	  *(sQWORD*)ptr->data = *mmx_register_values[operand->reg];
	  return true;
	  
    case o_crreg :				//Control
	  if (ctrl_conv[operand->reg] != 0xFF)
	    {*(sDEFAULT*)ptr->data = *control_register_values[ctrl_conv[operand->reg]];
	     return true;}
	  else return false;

    case o_dbreg :				//Debug
	  *(sDEFAULT*)ptr->data = *debug_register_values[dbg_conv[operand->reg]];
	  return true;}
 return false;}



bool	get_operand_address(op_t *operand, operand_properties *ptr, fixed_cache *cache, sBYTE mask)
{sADDR base, addr;
 sDEFAULT size;

 switch (operand->type)
   {case o_reg :
      if (operand->reg <= R_di)			//general registers, eax - edi
        {ptr->addr = general_register_values[operand->reg];
		 return true;}

	  if (operand->reg >= R_al && operand->reg <= R_bl)		//general regs al - bl
	    {ptr->addr = general_register_values[operand->reg - R_al];
		 return true;}

	  if (operand->reg >= R_ah && operand->reg <= R_bh)		//general regs ah - bh
	    {ptr->addr = (sBYTE*)(general_register_values[operand->reg - R_ah]) + 1;
		 return true;}

	  if (operand->reg >= R_es && operand->reg <= R_gs)		//segment regs
	    {ptr->addr = segment_register_values[operand->reg - R_es];
	     return true;}
       return false;										//unknown register
    
    case o_mem :
    case o_phrase :
    case o_displ :
	   base = get_memref_base(operand);
	   if (base == sADDR(-1)) return false;

       size = sizes_table[operand->dtyp];
	   if (!size) return false;

       addr = calculate_addr(operand) + base;
	   if (mask & OPERAND_ADDR)
	     {ptr->addr = (void*)addr;
	      return true;}
	
	   else if (mask & OPERAND_ADDRt)
	     {if (!MemoryExists(addr, size)) return false;
		  cache->addr_in_base = addr;
	      cache->size = size;
	      ptr->addr = cache->data;
	      return true;}

	   else if (mask & OPERAND_ADDRv)
	     {if (!MemoryExists(addr, size)) return false;
		  cache->addr_in_base = addr;
	      cache->size = size;
	      get_many_bytes(addr, cache->data, size);
	      ptr->addr = cache->data;
	      return true;}

    case o_imm :
	case o_near :
	case o_far :
	  return false;
	
	case o_fpreg :				//FPU
	  ptr->addr = fpu_register_values[operand->reg];
	  return true;
 
    case o_mmxreg :				//MMX
	  ptr->addr = mmx_register_values[operand->reg];
	  return true;
	
    case o_crreg :				//Control
	  if (ctrl_conv[operand->reg] != 0xFF)
	    {ptr->addr = control_register_values[ctrl_conv[operand->reg]];
	     return true;}
	  else return false; 

    case o_dbreg :				//Debug
	  ptr->addr = debug_register_values[dbg_conv[operand->reg]];
	  return true;}

return false;}




int		get_operand_address_ch(op_t *operand, operand_properties *ptr, fixed_cache *cache, sBYTE mask)
{sADDR base, addr, bp_addr, bp_limit;
 sDEFAULT size;
 int i;
	
 switch (operand->type)
   {case o_reg :
      if (operand->reg <= R_di)			//general registers, eax - edi
        {ptr->addr = general_register_values[operand->reg];
		 return true;}

	  if (operand->reg >= R_al && operand->reg <= R_bl)		//general regs al - bl
	    {ptr->addr = general_register_values[operand->reg - R_al];
		 return true;}

	  if (operand->reg >= R_ah && operand->reg <= R_bh)		//general regs ah - bh
	    {ptr->addr = (sBYTE*)(general_register_values[operand->reg - R_ah]) + 1;
		 return true;}

	  if (operand->reg >= R_es && operand->reg <= R_gs)		//segment regs
	    {ptr->addr = segment_register_values[operand->reg - R_es];
	     return true;}
       return false;										//unknown register
    
    case o_mem :
    case o_phrase :
    case o_displ :
	   base = get_memref_base(operand);
	   if (base == sADDR(-1)) return false;

       size = sizes_table[operand->dtyp];
	   if (!size) return false;

       addr = calculate_addr(operand) + base;
	  
	   if (mask & OPERAND_ADDR)
	     {ptr->addr = (void*)addr;
	      return true;}
	
	   else if (mask&OPERAND_ADDRt || mask&OPERAND_ADDRv)
	    {for (i = 0; i < context.breakpoints_number; i++)
           {bp_addr = context.breakpoints[i].addr;
		    bp_limit = bp_addr + context.breakpoints[i].size;
	     
		    if (addr >= bp_addr &&
				addr < bp_limit &&
				context.breakpoints[i].type&mask)
		    return EMUL_STATUS_BREAKPOINT | i;}
	   
	      if (!MemoryExists(addr, size)) return false;
		  cache->addr_in_base = addr;
	      cache->size = size;
	      ptr->addr = cache->data;
	      
		  if (mask&OPERAND_ADDRv)
		    get_many_bytes(addr, cache->data, size);

 	      return true;}
 

    case o_imm :
	case o_near :
	case o_far :
	  return false;
	
    case o_fpreg :				//FPU
	  ptr->addr = fpu_register_values[operand->reg];
	  return true;
 
    case o_mmxreg :				//MMX
	  ptr->addr = mmx_register_values[operand->reg];
	  return true;
	
    case o_crreg :				//Control
	  if (ctrl_conv[operand->reg] != 0xFF)
	    {ptr->addr = control_register_values[ctrl_conv[operand->reg]];
	     return true;}
	  else return false; 

    case o_dbreg :				//Debug
	  ptr->addr = debug_register_values[dbg_conv[operand->reg]];
	  return true;}
 return false;}



bool	get_operand_size(op_t *operand, operand_properties *ptr)
{if (sizes[operand->dtyp] != 0xFF)
   {ptr->size = sizes[operand->dtyp];
	return true;}
 else return false;}


void	flush_operands()
{int i;

 for (i = 0; i < MAX_OPERANDS_NUMBER; i++)
   if (op_cache[i].addr_in_base)
	 {patch_many_bytes(op_cache[i].addr_in_base, op_cache[i].data, op_cache[i].size);
      op_cache[i].addr_in_base = 0;}}



sADDR	get_memref_base(op_t *operand)
{int i;

 for (i = 0; i < sizeof(selectors) / sizeof(selector_associate); i++)
   if (*segment_register_values[operand->segrg - R_es] == selectors[i].real_selector)
     return selectors[i].base_address;



 return sADDR(-1);}





bool	MemoryExists(sADDR addr, sDEFAULT size)
{int i = 0;

 for (i = 0; i < size; i++)
   if (!isEnabled(addr + i)) return false;

 return true;}



sDEFAULT	calculate_addr(op_t *operand)
{sDEFAULT result;

 if (operand->specflag1 == 1 && operand->specflag2 != 0x24)
   {switch (operand->type)
      {case o_mem :
	     result = (*general_register_values[(operand->specflag2 >> 3) & 0x7] * (1 << (operand->specflag2 >> 6)));
	     break;
   
       case o_phrase :
	   case o_displ :
         result = *general_register_values[operand->specflag2 & 0x7] +
	              (*general_register_values[(operand->specflag2 >> 3) & 0x7] * (1 << (operand->specflag2 >> 6)));
	     break;}}
 
 else
   {switch (operand->type)
     {case o_mem :
        result = operand->addr;
        break;
 
      case o_phrase :
        result = *general_register_values[operand->phrase];
	    break;

	  case o_displ :
	    result = *general_register_values[operand->phrase] + operand->addr;
	    break;}}

 return result;}






void	create_stack_segment()
{sel_t selector;
 segment_t *segment1, *segment2;
 int i;

 if (stack_segment_base)
   return;

 selector = allocate_selector(0);

 segment1 = getseg(0x120000);
 segment2 = getseg(0xFFF00000);

 if (segment1)
   {if (!strcmp(get_segm_name(segment1), "STACK") && !strcmp(get_segm_class(segment1), "stack"))
	  {stack_segment_base = 0x120000;
       return;}}

 if (segment2)
   {if (!strcmp(get_segm_name(segment2), "STACK") && !strcmp(get_segm_class(segment2), "stack"))
	  {stack_segment_base = 0xFFF00000;
       return;}}

 if (!segment1)
   stack_segment_base = 0x120000;
 else if (!segment2)
   stack_segment_base = 0xFFF00000;
 else
   {msg("Cannot create stack segment\n");
    return;}

 if (!add_segm(selector, stack_segment_base, stack_segment_base + 0x10000, "STACK", "stack"))
   msg("Cannot create stack segment\n");
 else
   {segment1 = get_segm_by_sel(selector);
    set_segm_addressing(segment1, 1);
    for (i = stack_segment_base; i < stack_segment_base + 0x10000; i += 4)
      patch_long(i, 0);}}



void	create_peb_segment()
{sel_t selector;
 segment_t *segment;
 int i;

 if (peb_segment_base)
   return;

 segment = getseg(0x7FFDF000);

 if (segment)
   {if (!strcmp(get_segm_name(segment), "PEB") && !strcmp(get_segm_class(segment), "public"))
	  {peb_segment_base = 0x7FFDF000;
       return;}}

 if (!segment)
   peb_segment_base = 0x7FFDF000;
 else
   {msg("Cannot peb stack segment\n");
    return;}

 selector = allocate_selector(peb_segment_base>>4);
 if (!add_segm(selector, peb_segment_base, peb_segment_base + 0x1000, "PEB", "public"))
   msg("Cannot create peb segment\n");
 else
   {segment = get_segm_by_sel(selector);
    set_segm_addressing(segment, 1);
    for (i = peb_segment_base; i < (peb_segment_base + 0x1000); i += 4)
      patch_long(i, 0);}}





void	delete_stack_segment()
{del_segm(stack_segment_base, 1);
 stack_segment_base = 0;}


void	delete_peb_segment()
{del_segm(peb_segment_base, 1);
 peb_segment_base = 0;}




void change_eip()
{context.R_ip = get_screen_ea();}

void	change_cursor_pos()
{jumpto(context.R_ip);}

sADDR	get_cursor_pos()
{return get_screen_ea();}

void	idb_read(unsigned __int32 addr, void *buff, unsigned __int32 size)
{get_many_bytes(addr, buff, size);}

void	idb_write(unsigned __int32 addr, void *buff, unsigned __int32 size)
{patch_many_bytes(addr, buff, size);}

bool	idb_valid(unsigned __int32 addr, unsigned __int32 size)
{unsigned __int32 i;

 for (i = 0; i < size; i++)
   if (!isLoaded(addr + i))
     return false;

  return true;}


int item_size(sADDR addr)
{return get_item_size(addr);}


void	mem_free()
{if (indexed_IT != 0) LocalFree(indexed_IT);}

void QXZeroMemory(void *mem, unsigned long length)
{unsigned long i = length >> 3;
 while (i-- > 0) {*(unsigned __int64*)mem = 0;
                mem = (unsigned __int64*)mem + 1;}
 i = length & 0x3;
 while (i-- > 0) {*(unsigned char*)mem = 0;
                mem = (unsigned char*)mem + 1;}}

void QXmemcpy(void *dest, void *src, unsigned long length)
{unsigned long i = length >> 3;
 while (i-- > 0) {*(unsigned __int64*)dest = *(unsigned __int64*)src;
                dest = (unsigned __int64*)dest + 1;
                src = (unsigned __int64*)src + 1;}
 i = length & 0x3;
 while (i-- > 0) {*(unsigned char*)dest = *(unsigned char*)src;
                dest = (unsigned char*)dest + 1;
                src = (unsigned char*)src + 1;}}