#include <ida.hpp>
#include <idp.hpp>
#include <bytes.hpp>
#include <loader.hpp>
#include <kernwin.hpp>
#include <allins.hpp>

#ifdef _IDA_460
#include "ida460.h"
#else
#include "ida450.h"
#endif

//================================================================================//
//									Types										  //
//================================================================================//


typedef	sDEFAULT (*instr_handler)();



//================================================================================//
//									Constants									  //
//================================================================================//

	//==================== Operands properties ====================//
#define	MAX_OPERANDS_NUMBER			3

#define OPERAND_VALUE		0x01
#define	OPERAND_ADDRv		0x02
#define OPERAND_ADDRt		0x04
#define OPERAND_ADDR		0x08
#define	OPERAND_SIZE		0x10
#define OPERAND_address		(OPERAND_ADDRv | OPERAND_ADDRt | OPERAND_ADDR)


#define	OPERAND_READ		0x20
#define	OPERAND_WRITE		0x40


#define INSTRUCTIONS_TABLE_SIZE			1154

//================================================================================//
//									Structures									  //
//================================================================================//

/* This structure need for associate values in segment registers(in emulator context) with
   data, loaded in IDA database */

struct selector_associate
{unsigned __int16		real_selector;
 unsigned __int32		base_address;};


/* Operand properties:
   - value
   - address(for registers this field is pointer to this register in emul_context
             structure, for memory - pointer to "cache", this cache flushes to 
			 IDA base after emulation)
   - size, a lot of instructions handlers need to initialize this field
             (0 = byte, 1 = word, 2 = dword and 3 is qword) */

struct operand_properties
{void*		addr;
 sDEFAULT	size;
 char		data[64];};

 /* Cache structure, fixed type.
    This structure initialized before emulating instruction, then
	data flush to IDA base */

struct fixed_cache
{sADDR		addr_in_base;
 sDEFAULT	size;
 char		data[64];};


 /* Instruction parametres structure
    this structure exists for each instruction, defined
	in include file allins.hpp */

struct instr_properties
{
 
 /* describes fields of operand_properties structure
    require by handler of this instruction and conditions
    for checking breakpoints(4-byte alignment) */
	
 sBYTE		mask[3];

 /* Type of instruction handler - one of
    int, fpu and mmx type handler */
 
 sBYTE		stack_use;

 /* Address of instruction handler */

 instr_handler		handler;};


 /* Additional instruction params - stack buffer,
    prefixes, etc. */

struct additional_instr_params
{

 /* instruction prefix */
 sDEFAULT		prefix;

 /* instruction size */
 sDEFAULT		instr_size;

 /* stack buffer */
 char		stack_buffer[64];

};