#include <windows.h>

//================================================================================//
//									Types										  //
//================================================================================//

typedef	unsigned __int8		sBYTE;
typedef unsigned __int16	sWORD;
typedef unsigned __int32	sDWORD;
typedef unsigned __int64	sQWORD;

#ifdef x86_64
typedef sQWORD				sADDR;
typedef sQWORD				sDEFAULT;

#else
typedef sDWORD				sADDR;
typedef sDWORD				sDEFAULT;

#endif
	
//================================================================================//
//									Constants									  //
//================================================================================//

#define	FLAG_C		0x1
#define	FLAG_P		0x4
#define FLAG_A		0x10
#define	FLAG_Z		0x40
#define	FLAG_S		0x80
#define	FLAG_T		0x100
#define	FLAG_I		0x200
#define	FLAG_D		0x400
#define	FLAG_O		0x800

	//==================== Constants for registers ====================//

#ifdef x86_64
#define	GENERAL_REGISTERS_NUMBER		16
#define CONTROL_REGISTERS_NUMBER		5
#else
#define	GENERAL_REGISTERS_NUMBER		8
#define CONTROL_REGISTERS_NUMBER		4
#endif

#define	FPU_REGISTERS_NUMBER			8
#define	MMX_REGISTERS_NUMBER			8
#define DEBUG_REGISTERS_NUMBER			6

	//==================== Constants for breakpoints ====================//

#define	MAX_BREAKPOINTS_NUMBER			32

#define	BREAKPOINT_EXECUTE				0x80
#define	BREAKPOINT_READ					0x20
#define	BREAKPOINT_WRITE				0x40



//================================================================================//
//									Structures									  //
//================================================================================//


	//==================== Structures for breakpoints support ====================//

struct breakpoint
{sADDR		addr;
 sDEFAULT	size;
 sDEFAULT	type;};

	//==================== MAIN CONTEXT STRUCTURE ====================//

struct x86_emul_context
{

 //General registers
sDEFAULT	R_ax;
sDEFAULT	R_cx;
sDEFAULT	R_dx;
sDEFAULT	R_bx;
sDEFAULT	R_sp;
sDEFAULT	R_bp;
sDEFAULT	R_si;
sDEFAULT	R_di;

 //Control registers
sDEFAULT	R_flags;
sDEFAULT	R_ip;

 //FPU and MMX registers
union {long double	R_st0;
	   sQWORD		R_MM0;};

union {long double	R_st1;
	   sQWORD		R_MM1;};

union {long double	R_st2;
	   sQWORD		R_MM2;};

union {long double	R_st3;
	   sQWORD		R_MM3;};

union {long double	R_st4;
	   sQWORD		R_MM4;};

union {long double	R_st5;
	   sQWORD		R_MM5;};

union {long double	R_st6;
	   sQWORD		R_MM6;};

union {long double	R_st7;
	   sQWORD		R_MM7;};

 //Segment registers
sWORD		R_es;
sWORD		R_cs;
sWORD		R_ss;
sWORD		R_ds;
sWORD		R_fs;
sWORD		R_gs;

 //Debug registers
sDEFAULT	R_dr0;
sDEFAULT	R_dr1;
sDEFAULT	R_dr2;
sDEFAULT	R_dr3;
sDEFAULT	R_dr6;
sDEFAULT	R_dr7;

 //Control registers
sDEFAULT	R_cr0;
sDEFAULT	R_cr2;
sDEFAULT	R_cr3;
sDEFAULT	R_cr4;

 //Task register
sWORD		R_tr;

  //breakpoints area
sDEFAULT	breakpoints_number;
breakpoint	breakpoints[32];

sADDR		sys_breakpoint_addr;};



//================================================================================//
//								Public functions								  //
//================================================================================//


/* Call during initialization of plugin
   Set all values in emul_context structure to default
   Create stack and PEB segments */
bool	reset_configuration();


void	mem_free();

/* Functions for woring
   with emulator window */
HWND	create_emulator_window();
void	show_emulator_window(HWND hWnd);
void	destroy_emulator_window(HWND hWnd);
void	refresh_all(HWND hWnd);


/* Functions for read and write to IDA base
   (redirects to read_many_bytes and write_many_bytes) */
void	idb_read(sADDR addr, void *buff, sDEFAULT size);
void	idb_write(sADDR addr, void *buff, sDEFAULT size);
bool	idb_valid(sADDR addr, sDEFAULT size);

/* Emulator interface */
bool	emulate_instruction(sADDR addr);
int		emulate_instruction_ch(sADDR addr);

bool	trace_into(sADDR addr);
int		step_over(int unk0);
int		run_to_cursor(sADDR addr);
int		e_run(int unk0);
int		execute_till_return(int unk0);

void	change_eip();
void	change_cursor_pos();
sADDR	get_cursor_pos();


void	create_stack_segment();
void	create_peb_segment();
void	delete_stack_segment();
void	delete_peb_segment();


 /* STD */
void QXZeroMemory(void *mem, unsigned long length);
void QXmemcpy(void *dest, void *src, unsigned long length);
int item_size(sADDR addr);

extern "C" char *__cdecl _gcvt(double, int, char*);